#  legame/examples/seeking-demo.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Demonstrates seeking behavior of the MovingSprite class found in the
"sprite_enhancement" module.
"""
import argparse, logging
import pygame
from pygame import Surface
from pygame.draw import polygon, circle, line
from pygame.sprite import Sprite
from pygame.math import Vector2 as Vector
try:
	from pygame.locals import SRCALPHA, K_q, K_ESCAPE
except ImportError:
	from pygame import SRCALPHA, K_q, K_ESCAPE
from legame import vint
from legame.game import Game, GameState
from legame.sprite_enhancement import MovingSprite


class SeekingDemo(Game):
	"""
	Describe your game class here.
	"""

	fps	= 30

	def __init__ (self, options = None):
		self.set_resource_dir_from_file(__file__)
		Game.__init__(self, options)

	def initial_background(self, display_size):
		bg = Surface((400, 400))
		bg.fill((0,0,0))
		return bg

	def initial_state(self):
		return ChaseState()


class ChaseState(GameState):

	def __init__ (self):
		pygame.mouse.set_visible(False)
		cx, cy = Game.current.screen_rect.center
		mx, my = pygame.mouse.get_pos()
		Chaser(cx, cy, Mouse(mx, my))

	def key_down(self, event):
		"""
		Exit game immediately if K_ESCAPE key or "Q" key pressed
		"""
		if event.key in (K_ESCAPE, K_q):
			Game.current.shutdown()

	def quit_event(self, event):
		Game.current.shutdown()


class Mouse(MovingSprite, Sprite):

	height	= 18
	width	= 18

	def __init__(self, x, y):
		MovingSprite.__init__(self, x, y)
		Sprite.__init__(self, Game.current.sprites)
		Game.current.sprites.change_layer(self, Game.LAYER_ABOVE_PLAYER)
		self.color = pygame.Color("yellow")

	def update(self):
		self.position = Vector(pygame.mouse.get_pos())
		self.image = Surface(self.rect.size, SRCALPHA)
		circle(self.image, self.color, (9,9), 9, 1)
		line(self.image, self.color, (9,0), (9, 18))
		line(self.image, self.color, (0,9), (18, 9))
		self.update_rect()


class Chaser(MovingSprite, Sprite):

	height				= 24
	width				= 24
	max_turning_speed	= 10

	def __init__(self, x, y, mouse):
		MovingSprite.__init__(self, x, y)
		Sprite.__init__(self, Game.current.sprites)
		Game.current.sprites.change_layer(self, Game.LAYER_PLAYER)
		self.target = mouse
		self.color = pygame.Color("white")
		self.points = [
			(12, 0),
			(-12, 10),
			(-4, 0),
			(-12, -10)
		]
		self.center_point = Vector(12, 12)

	def update(self):
		self.turn_towards(self.target.position)
		self.image = Surface(self.rect.size)
		polygon(self.image, self.color, [vint(Vector(p).rotate(self.direction) + self.center_point) \
			for p in self.points])
		self.destination = self.target.position
		self.seek_motion()


if __name__ == '__main__':
	p = argparse.ArgumentParser()
	p.add_argument("--verbose", "-v", action = "store_true", help = "Show more detailed debug information")
	p.epilog = __doc__
	options = p.parse_args()
	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "[%(filename)24s:%(lineno)-4d] %(message)s"
	)

	p.exit(SeekingDemo().run())


#  end legame/examples/seeking-demo.py



#  legame/examples/board-game.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Demonstrates board game moves, jumps, state changes, and animations
"""
import argparse, logging
import random
try:
	from pygame.locals import K_q, K_ESCAPE
except ImportError:
	from pygame import K_q, K_ESCAPE
from pygame.sprite import Sprite
from legame.game import Game
from legame.board_game import BoardGame, GamePiece, BoardGameState
from legame.flipper import Flipper, FlipBetween, FlipThrough, FlipNone
from legame.exit_states import GSQuit


class TestGame(BoardGame):

	def __init__(self, options = None):
		self.set_resource_dir_from_file(__file__)
		BoardGame.__init__(self, options)
		self.my_color = random.choice(["r", "g", "b"])

	def initial_state(self):
		return GSSelect(cell = None)

# ----------------------------------------
# Game states:

class GSBase(BoardGameState):
	"""
	Used as the base class of all game states defined in this module.
	ESCAPE or Q button quits game.
	"""

	def key_down(self, event):
		"""
		Exit game immediately if K_ESCAPE key pressed
		"""
		if event.key in (K_ESCAPE, K_q):
			GSQuit()


class GSSelect(BoardGameState):
	"""
	Game state entered when the user must choose an empty space to fill, or their own block to move.
	attributes used:
		cell: the position that the last piece moved to, used as a reminder
	Example of changing state to this:
		GSSelect(cell = (x, y))
	"""

	may_click	= True	# May click on any block
	cell		= None

	def enter_state(self):
		Game.current.statusbar.write("GSSelect")
		self.reminder_timer = None if self.cell is None else Game.current.set_timeout(self.timeout, 4000)

	def click(self, cell, evt):
		if self.reminder_timer is not None:
			Game.current.clear_timeout(self.reminder_timer)
		if cell.is_mine():
			GSSelectMoveTarget(cell = cell)
		else:
			Block(cell, Game.current.my_color)
			self.cell = cell	# Used to jiggle the last piece moved
			self.reminder_timer = Game.current.set_timeout(self.timeout, 4000)

	def timeout(self):
		self.cell.piece().jiggle()

	def key_down(self, event):
		"""
		Exit game immediately if K_ESCAPE or "q" keys pressed
		"""
		if event.key in (K_ESCAPE, K_q):
			GSQuit()

	def exit_state(self, next_state):
		if self.reminder_timer is not None:
			Game.current.clear_timeout(self.reminder_timer)


class GSSelectMoveTarget(BoardGameState):
	"""
	Game state entered after the user choses their own block, setting up for a move.
	attributes used:
		cell: the position of the block to move
	Example of changing state to this:
		GSSelectMoveTarget(cell = (x, y))
	"""

	may_click	= True	# May click on any block
	cell		= None

	def enter_state(self):
		self.selected_piece = self.cell.piece().glow()
		Game.current.statusbar.write("GSSelectMoveTarget")

	def click(self, cell, evt):
		self.selected_piece.unglow()
		if cell.is_mine():
			self.selected_piece = cell.piece().glow()
		else:
			Game.current.play("jump.wav")
			self.selected_piece.move_to(cell, lambda: GSSelect(cell = cell))

	def key_down(self, event):
		"""
		Exit game immediately if K_ESCAPE key pressed
		"""
		if event.key in (K_ESCAPE, K_q):
			GSQuit()

# ----------------------------------------
# Sprites:

class Block(GamePiece, Flipper):

	def __init__(self, cell, color):
		self.image_folder = "Block/" + color
		GamePiece.__init__(self, cell, color)
		Flipper.__init__(self, FlipThrough("enter", fps = 25), FlipNone())
		Game.current.play("enter.wav")
		self.__glow = None

	def update(self):
		GamePiece.update(self)
		Flipper.update(self)

	def jiggle(self):
		self.flip(FlipBetween("jiggle", loops = 11, fps = 30), FlipNone())
		return self

	def glow(self):
		self.__glow = Glow(self.cell)
		return self

	def unglow(self):
		if self.__glow:
			self.__glow.kill()
			self.__glow = None
		return self

	def kill(self):
		Game.current.play("bust.wav")
		Sprite.kill(self)


class Glow(Flipper, Sprite):

	def __init__(self, cell, frame = 0):
		self.cell = cell
		self.rect = self.cell.rect()
		Sprite.__init__(self, Game.current.sprites)
		Game.current.sprites.change_layer(self, Game.LAYER_BELOW_PLAYER)
		Flipper.__init__(self, FlipBetween(loop_forever = True, frame = frame, fps = 30))


if __name__ == '__main__':
	p = argparse.ArgumentParser()
	p.add_argument("--quiet", "-q", action = "store_true",
		help = "Don't make sound")
	p.add_argument("--verbose", "-v", action = "store_true",
		help = "Show more detailed debug information")
	p.epilog = __doc__
	options = p.parse_args()
	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "[%(filename)24s:%(lineno)-4d] %(message)s"
	)

	p.exit(TestGame(options).run())


#  end legame/examples/board-game.py



#  legame/examples/steering-demo.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Simple demo which points an arrow at the mouse cursor.
"""
import argparse, logging
import pygame
from pygame import Surface
from pygame.draw import polygon, circle, line
from pygame.sprite import Sprite
from pygame.math import Vector2 as Vector
try:
	from pygame.locals import SRCALPHA, K_q, K_ESCAPE
except ImportError:
	from pygame import SRCALPHA, K_q, K_ESCAPE
from legame import vint
from legame.game import Game, GameState
from legame.sprite_enhancement import MovingSprite



class SteeringDemo(Game):
	"""
	Describe your game class here.
	"""

	fps	= 30

	def __init__ (self, options = None):
		self.set_resource_dir_from_file(__file__)
		Game.__init__(self, options)

	def initial_background(self, display_size):
		bg = Surface((400, 400))
		bg.fill((0,0,0))
		return bg

	def initial_state(self):
		return ChaseState()


class ChaseState(GameState):

	def __init__ (self):
		pygame.mouse.set_visible(False)
		cx, cy = Game.current.screen_rect.center
		mx, my = pygame.mouse.get_pos()
		Chaser(cx, cy, Mouse(mx, my))

	def key_down(self, event):
		"""
		Exit game immediately if K_ESCAPE key or "Q" key pressed
		"""
		if event.key in (K_ESCAPE, K_q):
			Game.current.shutdown()

	def quit_event(self, event):
		Game.current.shutdown()


class Mouse(MovingSprite, Sprite):

	height	= 18
	width	= 18

	def __init__(self, x, y):
		MovingSprite.__init__(self, x, y)
		Sprite.__init__(self, Game.current.sprites)
		Game.current.sprites.change_layer(self, Game.LAYER_ABOVE_PLAYER)
		self.color = pygame.Color("yellow")

	def update(self):
		self.position = Vector(pygame.mouse.get_pos())
		self.image = Surface(self.rect.size, SRCALPHA)
		circle(self.image, self.color, (9,9), 9, 1)
		line(self.image, self.color, (9,0), (9, 18))
		line(self.image, self.color, (0,9), (18, 9))
		self.update_rect()


class Chaser(MovingSprite, Sprite):

	height				= 24
	width				= 24
	max_turning_speed	= 6

	def __init__(self, x, y, mouse):
		MovingSprite.__init__(self, x, y)
		Sprite.__init__(self, Game.current.sprites)
		Game.current.sprites.change_layer(self, Game.LAYER_PLAYER)
		self.target = mouse
		self.color = pygame.Color("white")
		self.points = [
			(12, 0),
			(-12, 10),
			(-4, 0),
			(-12, -10)
		]
		self.center_point = Vector(12, 12)

	def update(self):
		self.turn_towards(self.target.position)
		self.image = Surface(self.rect.size)
		polygon(self.image, self.color, [vint(Vector(p).rotate(self.direction) + self.center_point) \
			for p in self.points])
		MovingSprite.update(self)


if __name__ == '__main__':
	p = argparse.ArgumentParser()
	p.add_argument("--verbose", "-v", action = "store_true", help = "Show more detailed debug information")
	p.epilog = __doc__
	options = p.parse_args()
	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "[%(filename)24s:%(lineno)-4d] %(message)s"
	)

	p.exit(SteeringDemo().run())


#  end legame/examples/steering-demo.py



#  legame/exit_states.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides GameState classes which exit a game after showing an animated message.
"""
from pygame.sprite import Sprite
from legame.game import Game, GameStateFinal


class GSExitAnimation(GameStateFinal):
	"""
	A game state which shows an animated message, which shuts down the game when complete.
	"""

	def enter_state(self):
		ExitAnimation(self.image)


class GSWon(GSExitAnimation):

	image = "you-win.png"


class GSLost(GSExitAnimation):

	image = "game-over.png"


class GSQuit(GSExitAnimation):

	image = "bye-bye.png"


class ExitAnimation(Sprite):
	"""
	An animated sprite which drops down from the top of the display, and closes the
	current game when the animation is complete.
	"""

	anim_duration		= 0.33
	game_over_delay		= 777
	exit_timer			= None

	def __init__(self, image):
		Sprite.__init__(self, Game.current.sprites)
		Game.current.sprites.change_layer(self, Game.LAYER_OVERLAY)
		self.image = Game.current.resources.image(image)
		self.rect = self.image.get_rect()
		self.rect.top = -self.rect.height
		self.rect.centerx = Game.current.screen_rect.centerx
		self.__frame_step = (Game.current.screen_rect.centery - self.rect.centery) \
			// self.anim_duration // Game.current.fps

	def update(self):
		if self.exit_timer is None:
			centery = Game.current.background.get_rect().centery
			if self.rect.centery < centery:
				self.rect.centery = min(self.rect.centery + self.__frame_step, centery)
			else:
				self.exit_timer = Game.current.set_timeout(Game.current.shutdown, self.game_over_delay)


#  end legame/exit_states.py



#  legame/neighbors.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides a means for dividing up the screen into sub-sections within which MovingSprites are
cross-checked for nearness to each other. This allows MovingSprite instances to modify their
behavior on the basis of their proximity to other MovingSprites, without having to cross-check
every instance on the screen against every other instance.
"""
from math import floor


class Neighborhood:
	"""
	Divides an area of the screen into "Quadrant"s which are used to determine
	which sprites (or any thing really) are nearby which others.

	Quadrants are a series of squares which overlap on both the x, y axes. This
	reduces the search space when determining whether two things need to be
	notified of each other's existence.

	This can best be described visually. This is a representation of an
	Neighborhood covering a 4x3 grid. It spans 3 Quadrants on the x axis, (labeled
	0, 1, 2) and 2 quadrants on the y axis (labeled 0, 1). Inside of each cell in
	the illustration is a list of the x,y coordinates of all of the the quadrants
	which overlap that cell:

		---------------------
		|         0         |
		          ---------------------
		          |         1         |
		                    ---------------------
		                    |         2         |
		-----------------------------------------  ---
		| 0,0     | 0,0     |     2,0 |     2,0 |     |
		|         |   1,0   |   1,0   |         |     |
		|         |         |         |         |     |
		|         |         |         |         |     |
		-----------------------------------------  0  | ---
		| 0,0     | 0,0     |     2,0 |     2,0 |     |    |
		|         |   1,0   |   1,0   |         |     |    |
		|         |   1,1   |   1,1   |         |     |    |
		| 0,1     | 0,1     |     2,1 |     2,1 |     |    |
		-----------------------------------------  ---   1 |
		|         |         |         |         |          |
		|         |         |         |         |          |
		|         |   1,1   |   1,1   |         |          |
		| 0,1     | 0,1     |     2,1 |     2,1 |          |
		-----------------------------------------  --------

	Notice that each quadrant takes up four cells, and each cell shares at least
	one quadrant with all the cells surrounding it (above-left, above, above-right,
	left, right, below-left, below, below-right - 8 in total).

	Let's take a close look at the third cell from the left, third from the top.
	It is covered by two quadrants:
		1,1  2,1

	The cell immediately above it is in these quadrants:
		2,0  1,0  1,1  2,1
	... so these two cells are both in quadrant 1,1, and both in quadrant 2,1

	The cell above and to the left (second from left, second from top), is
	contained in these quadrants:
		0,0  1,0  1,1  0,1
	... so our cell (three cells from the left, three cells from the top), shares
	only quadrant 1,1

	The cell above and to the right (fourth from left, second from top), is
	contained in these quadrants:
		2,0  2,1
	... so our cell (three cells from the left, three cells from the top), shares
	only quadrant 2,1

	When an "observed" sprite is placed, it is given membership in the quadrants
	which contain the cell that it is in. Detecting interaction between sprites is
	done for every quadrant, and that reduces the amount of comparisons neccessary
	between sprites.

	To use a Neighborhood, instantiate the Neighborhood class, add the sprites you
	want notified of each other's existence using the "observe" function when they
	are instantiated, and call the "notify_sprites" function of the Neighborhood
	periodically. Typically this involves overriding the "_end_loop" function of
	the Game class and calling "notify_sprites" from there. Alternatively, you
	could call "notify_sprites" in a GameState's "loop_end" function. The only
	drawback to the second method is that you need to make sure that this call is
	made from every GameState which needs it.

	Don't forget to call the "ignore" function if/when a sprite is killed.
	Otherwise, it will remain in the list of observed sprites, and all the math
	necessary for keeping track of it will still be done. That will slow down
	your game. You might put a call to the "Neighborhood.ignore" function in the
	sprite's "kill" function.
	"""

	def __init__(self, rect, cells_x, cells_y):
		"""
		Set the area to be watched by this Neighborhood.
		(God, that sounds so Orwellian!)

		When constructing a Neighborhood, you must pass the following required
		arguments:

		rect (pygame.Rect): Identifies the area encompassing the Neighborhood.
							Only sprites in this rect will be notified of each
							others' existence. Normally,  this will be the
							Game.screen_rect, or some similar play area.
		cells_x (int)	  : The number of "cells" to divide the Neighborhood into
							on the x-axis - not "quadrants". A "quadrant" spans 3
							cells across. See the docstring for this module for a
							full explanation.
		cells_y (int)	  : The number of "cells" to divide the Neighborhood into
							on the y-axis - not "quadrants". (See "cells_x" above.)
		"""

		# Initialize either height/width based on cells_x/cells_y, or cells_x/cells_y based on height/width:
		self.rect = rect
		self.cells_x = cells_x
		self.cells_y = cells_y
		self.cell_width = self.rect.width / cells_x
		self.cell_height = self.rect.height / cells_y

		# self._quadrants is a flat, 1-dimensional list of Quadrant objects.
		# Each quadrant has an .x and .y attribute, corresponding to the first "cell" it covers.
		self._quadrants = []

		# self.__quadrant_maps is a two-dimensional which contains all the Quadrant instances, ordered by
		# their x,y coordinate which is the coordinate of the top-left cell covered by the Quadrant
		# First indice is the "x" axis, second indice is the "y" axis
		self.__quadrant_maps = [ [ None for y in range(0, self.cells_y - 1) ] for x in range(self.cells_x - 1) ]
		for x in range(self.cells_x - 1):
			for y in range(0, self.cells_y - 1):
				self._quadrants.append(Quadrant(x, y))
				self.__quadrant_maps[x][y] = self._quadrants[len(self._quadrants) - 1]

		self.count = len(self._quadrants)

		# self.cell_lookup is a 2-dimensional list, each element of which will contain a list of
		# references to the Quadrant instances which overlap that cell:
		self.__cell2quad_maps = [ [ [] for y in range(self.cells_y) ] for x in range(self.cells_x) ]

		# In a 4x3 grid, these are the class attributes:
		# cells_x:		4	- which makes for 3 quadrants across
		# cells_y:		3	- which makes for 2 quadrants down
		# quadrants:	3-across, with "x" indexes of 0, 1, 2;      2-down, with "y" indexes of 0, 1
		# __cell2quad_maps:	4-across, with "x" indexes of 0, 1, 3, 3;   3-down, with "y" indexes of 0, 1, 2
		#
		# Populate lookup tables:
		for x in range(self.cells_x - 1):			# in a 4x3 grid, iterate through 0, 1, 2
			for y in range(self.cells_y - 1):		# in a 4x3 grid, iterate through 0, 1
				for span_x in range(2):				# iterate through 0, 1
					#if x + span_x == self.cells_x: continue
					for span_y in range(2):			# iterate through 0, 1
						#if y + span_y == self.cells_y: continue
						self.__cell2quad_maps[x + span_x][y + span_y].append(self.__quadrant_maps[x][y])

		# self._observed_sprites_list are a list of all sprites to keep track of in the area:
		self._observed_sprites_list = []

	@property
	def cells(self):
		"""
		Returns a cell-to-quadrants list - really only should be used for testing.
		"""
		return self.__cell2quad_maps

	@property
	def all_quadrants(self):
		"""
		Returns the list of Quadrant objects - really only should be used for testing.
		"""
		return self._quadrants

	def sprites_in(self, x, y):
		"""
		Returns a list of sprites which occupy the quadrant specified by the given x/y coordinates.
		"""
		return self.__quadrant_maps[x][y].sprites

	def quadrant(self, x, y):
		"""
		Returns the single quadrant whose "top-left cell" occupies the "cell" x/y position given.
		"""
		return self.__quadrant_maps[x][y]

	def observe(self, sprite):
		"""
		Add a sprite to observe.
		You can add the same sprite more than once. All it will do is slow down your game.
		Don't do that.
		"""
		self._observed_sprites_list.append(sprite)

	def ignore(self, sprite):
		"""
		Remove a sprite from the list of sprites to observe.
		Hopefully, you haven't added it twice, because if so, it'll still be here.
		"""
		self._observed_sprites_list.remove(sprite)

	def notify_sprites(self):
		"""
		Re-calculate quadrant membership of all the sprites observed and notify the observed sprites
		when another observed sprite is within the same quadrant.
		"""
		for quadrant in self._quadrants:
			quadrant.sprites.clear()
		for sprite in self._observed_sprites_list:
			try:
				for quadrant in self.__cell2quad_maps[floor(sprite.x / self.cell_width)][floor(sprite.y / self.cell_height)]:
					quadrant.sprites.append(sprite)
			except IndexError:
				pass
		for quadrant in self._quadrants:
			cnt = len(quadrant.sprites)
			if cnt > 1:
				for a in range(cnt - 1):
					for b in range(a + 1, cnt):
						quadrant.sprites[a].notice(quadrant.sprites[b])
						quadrant.sprites[b].notice(quadrant.sprites[a])


class Quadrant:
	"""
	Division of an Neighborhood which covers at most 9 "cells".
	"""

	def __init__(self, x, y):
		self.x = x			# These values (and this whole class, really) are only
		self.y = y			# Used for debugging. The meat of this class is "sprites".
		self.sprites = []	# A list of sprites which are determined to be in this Quadrant.


class Neighbor:
	"""
	Demonstrates an implementation of the "notice" function, which is called when two things are within
	the same quadrant.
	"""

	def notice(self, neighbor):
		"""
		Informs this sprite that there's another sprite within a quadrant distance, allowing it
		"decide" what to do with the neighboring sprite.
		"""


#  end legame/neighbors.py



#  legame/joiner.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides the BroadcastJoiner class, a subclass of Dialog and BroadcastConnector
which returns a Messenger object already connected to another player over the
network.
"""
import importlib, re, threading
from time import time
from socket import socket, AF_INET, SOCK_DGRAM
from pgdialog import 	Dialog, Button, Label, Textbox, Radio, HorizontalLayout, \
						ALIGN_LEFT, ALIGN_CENTER
from cable_car.broadcast_connector import BroadcastConnector
from cable_car.direct_connect import DirectClient, DirectServer
from cable_car.messenger import Messenger


def get_my_ip():
	sock = socket(AF_INET, SOCK_DGRAM)
	sock.connect(('8.8.8.8', 7))
	return sock.getsockname()[0]


class JoinerDialog(Dialog):
	"""
	Base class of dialog which allows the user to connect to another player over the network.
	"""

	transport					= "json"
	caption						= "Select a game to join"
	background_color			= (20,20,80)
	background_color_disabled	= (0,0,40)
	foreground_color			= (180,180,255)
	foreground_color_hover		= (220,220,255)
	shutdown_delay				= 0.5

	def __init__(self, options = None):
		"""
		The "options" argument is expected to be a dictionary, the items of which are
		set as attributes of the game during initialization. Some appropriate key/value
		pairs to pass to the __init__ function would be:

			client
			server
			tcp_port
			udp_port
			transport

		"""
		if options is not None:
			for varname, value in options.__dict__.items():
				setattr(self, varname, value)
		Dialog.__init__(self)
		module = importlib.import_module("cable_car.%s_messages" % self.transport)
		globals().update({ name: module.__dict__[name] for name in module.__dict__})
		Message.register_messages()
		self.statusbar = Label(
			"",
			align = ALIGN_LEFT,
			font_size = 15,
			width = 540,
			foreground_color = self.foreground_color,
			background_color = self.background_color
		)

	def close(self):
		"""
		Override Dialog.close()
		Turn off "_connect_enable" so that connection threads exit.
		Sets up a pause before closing when "shutdown_delay" is not zero.
		"""
		self._connect_enable = False
		self._quitting_time = time() + self.shutdown_delay
		self.loop_end = self._closing

	def _closing(self):
		"""
		Function which replaces "loop_end" when closing JoinerDialog.
		Inserts a pause before killing the main loop, closing the dialog.
		"""
		if time() >= self._quitting_time:
			self._run_loop = False


class BroadcastJoiner(JoinerDialog, BroadcastConnector):
	"""
	A dialog which allows the user to connect to another player over the network by
	sending UDP broadcast messages and listening for connection attempts from
	machines which have received broadcast packets.

	When a connection is made, user and host name is sent to the remote computer.
	This is displayed on the buttons on this dialog, and the user can select one of
	the remote computers by clicking on one of the buttons.

	This class uses the cable_car Messenger class which allows for two different
	types of transport to be used; "json" or "byte". The "json" transport is a lot
	easier to implement, but requires more network bandwidth and may be slow for
	very busy network games. In contrast, the "byte" transport is very lightweight,
	but requires that you write message encoding and decoding routines yourself.

	The default transport is "json", but this can be overridden by defining a class
	variable in a subclass, or by passing "transport = byte" as an option to the
	__init__ function.

	"""

	def __init__(self, options = None):
		JoinerDialog.__init__(self, options)
		self.messengers = []
		self.messenger = None
		self._address_buttons = []
		for idx in range(4):
			button = Button(
				"",
				font_size = 20,
				padding = 12,
				margin = (2,10),
				foreground_color = self.foreground_color,
				foreground_color_hover = self.foreground_color_hover,
				background_color_disabled = self.background_color_disabled,
				background_color = self.background_color,
				click_handler = self._button_click,
				disabled = True,
				index = idx
			)
			self._address_buttons.append(button)
			self.append(button)
		self._address_buttons[0].margin_top = 10
		self.statusbar.text = "Waiting for other players to appear on the network"
		self.append(self.statusbar)
		self.on_connect_function = self.on_connect

	def show(self):
		self.initialize_display()
		self._start_connector_threads()
		self._main_loop()
		self._connect_enable = False
		self.join_threads()

	def on_connect(self, sock):
		"""
		Called when a socket connection is made, creates a new Messenger from new
		socket and appends it to the "messengers" list
		"""
		msgr = Messenger(sock, self.transport)
		# These properties of the Messenger are only relevant when the user
		# is selecting one of many NetworkMessengers, hence are only defined here.
		msgr.remote_user = None
		msgr.remote_hostname = None
		msgr.id_sent = False
		msgr.id_received = False
		msgr.was_invited = False
		msgr.invited_me = False
		msgr.accepted_invitation = False
		self.messengers.append(msgr)

	def loop_end(self):
		"""
		Function called each time through Dialog._main_loop(), updates the buttons,
		statusbar, and messengers.
		"""

		if self._connect_enable:
			# Determine what to display on the associated button:
			for idx in range(len(self.messengers)):
				msgr = self.messengers[idx]
				button = self._address_buttons[idx]
				if msgr.closed:
					button.text = "Connection to %s closed" % (msgr.remote_ip)
					button.disabled = True
					continue
				if not msgr.id_sent:
					button.text = "Connected to %s" % (msgr.remote_ip)
					msgr.send(MsgIdentify())
					msgr.id_sent = True
				msgr.xfer()
				# Handle responses from this Messenger
				message = msgr.get()
				if message is not None:
					if isinstance(message, MsgIdentify):
						msgr.id_received = True
						msgr.remote_hostname = message.hostname
						msgr.remote_user = message.username
						button.text = "%s on %s (click to invite)" % (msgr.remote_user, msgr.remote_hostname)
						button.disabled = False
					elif isinstance(message, MsgJoin):
						if msgr.was_invited:
							# Player which was invited accepted invitation - select
							button.text = "%s on %s accepted!" % (msgr.remote_user, msgr.remote_hostname)
							self._select(msgr)
						else:
							msgr.invited_me = True
							button.text = "%s on %s wants to play (click to accept)" % (msgr.remote_user, msgr.remote_hostname)
					else:
						raise Exception("Messenger received an unexpected message: " + message.__class__.__name__)

		else:
			if self._udp_broadcast_exc is None \
				and self._udp_listen_exc is None \
				and self._tcp_listen_exc is None:
				self.statusbar.text = "Connected." if self.messenger else "Cancelled."
			else:
				errors = []
				if self._udp_broadcast_exc: errors.append("Broadcast: " + self._udp_broadcast_exc.__str__())
				if self._udp_listen_exc: errors.append("Listen: " + self._udp_listen_exc.__str__())
				if self._tcp_listen_exc: errors.append("Socket: " + self._tcp_listen_exc.__str__())
				self.statusbar.text = ", ".join(errors)
			self.close()

	def _button_click(self, button):
		"""
		Click event fired when one of the address listings is clicked.
		"""
		msgr = self.messengers[button.index]
		if msgr.invited_me:
			msgr.send(MsgJoin())
			self._select(msgr)
		elif not msgr.was_invited:
			msgr.send(MsgJoin())
			msgr.was_invited = True
			button.text = "Waiting for %s on %s to accept" % (msgr.remote_user, msgr.remote_hostname)
			button.disabled = True

	def _select(self, messenger):
		"""
		Called when accepting an invitation or the other player accepts an invitation.
		Sets the "messenger" and exits the game joiner.
		"""
		self.messenger = messenger
		for msgr in self.messengers:
			if msgr is not self.messenger:
				msgr.close()
		self.close()


class DirectJoiner(JoinerDialog):

	tcp_port			= 8223		# Port to connect to / listen on
	transport			= "json"	# cable_car transport to use.
	timeout				= None		# Absolute limit on how long to wait.

	__connector			= None		# Instance of DirectClient or DirectServer
	__connect_thread	= None		# Connector thread
	__connect_exc		= None		# Exception raised in self.__connector thread
	__connect_failed	= False

	def __init__(self, options = None):
		JoinerDialog.__init__(self, options)
		self.messenger = None
		self.append(HorizontalLayout(
			Radio("mode", "Client",
				foreground_color = self.foreground_color,
				foreground_color_hover = self.foreground_color_hover,
				background_color_disabled = self.background_color_disabled,
				background_color = self.background_color,
				click_handler = self.mode_select
			),
			Radio("mode", "Server",
				foreground_color = self.foreground_color,
				foreground_color_hover = self.foreground_color_hover,
				background_color_disabled = self.background_color_disabled,
				background_color = self.background_color,
				click_handler = self.mode_select
			)
		))
		self.ip_entry = Textbox(get_my_ip(),
			font_size = 32,
			disabled = True,
			align = ALIGN_CENTER
		)
		self.append(self.ip_entry)
		self.start_button = Button("Connect",
			disabled = True,
			foreground_color = self.foreground_color,
			foreground_color_hover = self.foreground_color_hover,
			background_color_disabled = self.background_color_disabled,
			background_color = self.background_color,
			click_handler = self.start
		)
		self.append(self.start_button)
		self.statusbar.text = "Select the mode (client or server)"
		self.append(self.statusbar)

	def show(self):
		self.initialize_display()
		self._main_loop()
		self._connect_enable = False
		if self.__connect_thread: self.__connect_thread.join()

	def mode_select(self, radio):
		if radio.text == "Client":
			self.ip_entry.enabled = True
			self.start_button.enabled = re.match("^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$", self.ip_entry.text)
		else:
			self.ip_entry.enabled = False
			self.start_button.enabled = True
			self.ip_entry.text = get_my_ip()

	def start(self, pos):
		if Radio.selected_value("mode") == "Client":
			if re.match("^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$", self.ip_entry.text):
				self.disable_widgets()
				self.__connect_thread = threading.Thread(target = self.client_connect)
				self.statusbar.text = "Connecting to %s ..." % self.ip_entry.text
			else:
				self.ip_entry.focus()
				self.statusbar.text = "Enter a valid ip address"
				return
		else:
			self.disable_widgets()
			self.__connect_thread = threading.Thread(target = self.server_connect)
			self.statusbar.text = "Listening for client connection ..."
		self.__connect_thread.start()

	def disable_widgets(self):
		for widget in self.widgets(): widget.disabled = True

	def client_connect(self):
		self.__connect(DirectClient(self.tcp_port, self.ip_entry.text))

	def server_connect(self):
		self.__connect(DirectServer(self.tcp_port, self.ip_entry.text))

	def __connect(self, connector):
		self.__connector = connector
		self.__connector.timeout = self.timeout
		self.__connector.connect()
		if self.__connector.socket is None:
			self.__connect_failed = True
		else:
			self.messenger = Messenger(self.__connector.socket, self.transport)
			self.messenger.id_sent = False
			self.messenger.id_received = False

	def loop_end(self):
		"""
		Function called each time through Dialog._main_loop()
		"""
		if self.messenger:
			if self.messenger.closed:
				self.statusbar.text = "Connection to %s closed" % (self.messenger.remote_ip)
			else:
				if self.messenger.id_sent and self.messenger.id_received:
					self.statusbar.text = "%s at %s connected" % (self.messenger.remote_user, self.messenger.remote_hostname)
					self.close()
				elif not self.messenger.id_sent:
					self.statusbar.text = "Connected to %s" % (self.messenger.remote_ip)
					self.messenger.send(MsgIdentify())
					self.messenger.id_sent = True
				self.messenger.xfer()
				message = self.messenger.get()
				if message is not None:
					if isinstance(message, MsgIdentify):
						self.messenger.id_received = True
						self.messenger.remote_hostname = message.hostname
						self.messenger.remote_user = message.username
					else:
						self.__connector.cancel()
						err = "Messenger received an unexpected message: " + message.__class__.__name__
						logging.error(err)
						self.statusbar.text = err
		elif self.__connect_failed:
			# Only happens when connector thread faults out.
			self.close()


if __name__ == '__main__':
	import argparse, logging

	p = argparse.ArgumentParser()
	p.add_argument("--transport", type = str, default = "json")
	p.add_argument("--verbose", "-v", action = "store_true", help = "Show more detailed debug information")
	p.add_argument("--udp-port", type = int, default = 8222)
	p.add_argument("--tcp-port", type = int, default = 8223)
	p.add_argument("--direct", "-d", action = "store_true", help = "Connect by ip address instead of using udp broadcast discovery.")
	options = p.parse_args()

	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "%(relativeCreated)6d [%(filename)24s:%(lineno)-4d] %(message)s"
	)

	if options.direct:

		joiner = DirectJoiner(options)
		joiner.shutdown_delay = 0.5
		joiner.timeout = 15.0
		joiner.show()

	else:

		joiner = BroadcastJoiner(options)
		joiner.shutdown_delay = 0.5
		joiner.timeout = 15.0
		joiner.show()

		print("Addresses:")
		print(joiner.addresses())
		print("Messengers:")
		print([messenger.remote_ip for messenger in joiner.messengers])

	print("Selected:")
	print(joiner.messenger.remote_ip if joiner.messenger else None)


#  end legame/joiner.py



#  legame/tests/resources_test.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
import pytest, os, pygame
import legame
from legame.game import Game
from legame.resources import Resources, ImageSet
from pygame import Surface

@pytest.fixture(autouse = True)
def resdump():
	"""
	Returns Resources with "resource_dump" True
	"""
	return FakeGame().dump_resources

@pytest.fixture(autouse = True)
def res():
	"""
	Returns Resources with "resource_dump" False
	"""
	return FakeGame().resources


class FakeGame:
	def __init__(self):
		pygame.mixer.init()
		Game.current = self
		examples = os.path.join(os.path.dirname(os.path.dirname(legame.__file__)), "examples")
		if not os.path.isdir(examples):
			raise NotADirectoryError(examples)
		self.resource_dir = os.path.join(os.path.abspath(examples), "resources")
		if not os.path.isdir(self.resource_dir):
			raise NotADirectoryError(self.resource_dir)
		self.dump_resources = Resources(self.resource_dir, True)
		self.resources = Resources(self.resource_dir, False)

def test_img_path(resdump):
	assert os.path.isfile(resdump.image("Block/r/00.png"))

def test_img_load(res):
	img = res.image("Block/r/00.png", convert = False)
	assert isinstance(img, Surface)

def test_sound_path(resdump):
	assert os.path.isfile(resdump.sound("bust.wav"))

def test_sound_load(res):
	assert isinstance(res.sound("bust.wav"), pygame.mixer.Sound)

def test_image_set(resdump):
	imgset = resdump.image_set("Block")
	assert isinstance(imgset, ImageSet)
	assert imgset.count == 0

def test_subset(resdump):
	imgset = resdump.image_set("Block")
	imgset = imgset.variant("b")
	assert isinstance(imgset, ImageSet)
	assert imgset.count == 1
	assert imgset.last_index == 0
	imgset = imgset.variant("enter")
	assert isinstance(imgset, ImageSet)
	assert imgset.count == 3
	assert imgset.last_index == 2
	assert os.path.isfile(imgset.images[0])
	assert os.path.isfile(imgset.images[0])
	assert os.path.isfile(imgset.images[2])
	imgset = resdump.image_set("Block")
	imgset = imgset.variant("b/enter")
	assert isinstance(imgset, ImageSet)
	assert imgset.count == 3
	assert os.path.isfile(imgset.images[0])
	assert os.path.isfile(imgset.images[0])
	assert os.path.isfile(imgset.images[2])

def test_imgset_variant(resdump):
	imgset = resdump.image_set("Block")
	imgset = imgset.variant("b/enter")
	assert isinstance(imgset, ImageSet)
	assert imgset.count == 3
	assert os.path.isfile(imgset.images[0])
	assert os.path.isfile(imgset.images[0])
	assert os.path.isfile(imgset.images[2])

def test_imgset_from_subpath(resdump):
	imgset = resdump.image_set("Block/b/enter")
	assert isinstance(imgset, ImageSet)
	assert imgset.count == 3
	assert os.path.isfile(imgset.images[0])
	assert os.path.isfile(imgset.images[0])
	assert os.path.isfile(imgset.images[2])

#  legame/tests/resources_test.py
